#![warn(clippy::all)]

use std::env;

use serenity::{
    async_trait,
    model::{
        gateway::Ready,
        id::GuildId,
        interactions::{
            application_command::{ApplicationCommandOptionType, ApplicationCommandPermissionType},
            Interaction, InteractionResponseType,
        },
    },
    prelude::*,
};
use songbird::{tracks::TrackHandle, SerenityInit};

mod commands;

struct CurrentlyPlayingTrack;

impl TypeMapKey for CurrentlyPlayingTrack {
    type Value = Option<TrackHandle>;
}

struct CurrentVolume;

impl TypeMapKey for CurrentVolume {
    type Value = f32;
}

struct Handler;

#[async_trait]
impl EventHandler for Handler {
    async fn interaction_create(&self, mut ctx: Context, interaction: Interaction) {
        if let Interaction::ApplicationCommand(command) = interaction {
            let content = match command.data.name.as_str() {
                "join" => commands::join(&ctx, &command).await,
                "leave" => commands::leave(&ctx, &command).await,
                "play" => commands::play(&mut ctx, &command).await,
                "stop" => commands::stop(&ctx, &command).await,
                "volume" => commands::set_volume(&mut ctx, &command).await,
                "loop" => commands::set_loop(&mut ctx, &command).await,
                _ => "not implemented :(".to_string(),
            };

            if let Err(why) = command
                .create_interaction_response(&ctx.http, |response| {
                    response
                        .kind(InteractionResponseType::ChannelMessageWithSource)
                        .interaction_response_data(|message| message.content(content))
                })
                .await
            {
                println!("Cannot respond to slash command: {}", why);
            }
        }
    }

    async fn cache_ready(&self, ctx: Context, guilds: Vec<GuildId>) {
        for guild in guilds {
            let commands = guild
                .set_application_commands(&ctx.http, |commands| {
                    commands
                        .create_application_command(|command| {
                            command
                                .name("join")
                                .description("Join your current channel")
                                .default_permission(false)
                        })
                        .create_application_command(|command| {
                            command
                                .name("leave")
                                .description("Leave the bot's current channel")
                                .default_permission(false)
                        })
                        .create_application_command(|command| {
                            command
                                .name("play")
                                .description("Play a song")
                                .create_option(|option| {
                                    option
                                        .name("url")
                                        .description("The URL of the song to play")
                                        .kind(ApplicationCommandOptionType::String)
                                        .required(true)
                                })
                                .default_permission(false)
                        })
                        .create_application_command(|command| {
                            command
                                .name("stop")
                                .description("Stop any currently playing songs")
                                .default_permission(false)
                        })
                        .create_application_command(|command| {
                            command
                                .name("volume")
                                .description("Set the bot's playback volume")
                                .create_option(|option| {
                                    option
                                        .name("volume")
                                        .description("The volume on a scale from 0 to 100")
                                        .kind(ApplicationCommandOptionType::Number)
                                        .required(true)
                                })
                                .default_permission(false)
                        })
                        .create_application_command(|command| {
                            command
                                .name("loop")
                                .description("Enable or disable looping the currently playing track")
                                .create_option(|option| {
                                    option
                                        .name("loop")
                                        .description("Whether or not the track should loop")
                                        .kind(ApplicationCommandOptionType::Boolean)
                                        .required(true)
                                })
                                .default_permission(false)
                        })
                })
                .await
                .expect("Couldn't create commands");

            println!("I created the following guild commands: {:#?}", commands);

            let role_id = env::var("ROLE_ID")
                .expect("Expected a role id in the environment")
                .parse::<u64>()
                .expect("Role id is not a valid id");

            let permissions = guild
                .set_application_commands_permissions(&ctx.http, |permissions| {
                    for command in commands {
                        permissions.create_application_command(|permissions| {
                            permissions
                                .id(command.id.into())
                                .create_permissions(|permission| {
                                    permission
                                        .id(role_id)
                                        .kind(ApplicationCommandPermissionType::Role)
                                        .permission(true)
                                })
                        });
                    }
                    permissions
                })
                .await
                .expect("Couldn't set permissions");

            println!("I created the following permissions: {:#?}", permissions);
        }
    }

    async fn ready(&self, _ctx: Context, ready: Ready) {
        println!("{} is connected!", ready.user.name);
    }
}

#[tokio::main]
async fn main() {
    let token =
        env::var("DISCORD_TOKEN").expect("Expected a bot token in the environment: DISCORD_TOKEN");

    let application_id: u64 = env::var("APPLICATION_ID")
        .expect("Expected an application id in the environment: APPLICATION_ID")
        .parse()
        .expect("application id is not a valid id");

    env::var("ROLE_ID")
        .expect("Expected a role id in the environment: ROLE_ID")
        .parse::<u64>()
        .expect("Role id is not a valid id");

    env::var("CHANNEL_ID")
        .expect("Expected a channel id in the environment: CHANNEL_ID")
        .parse::<u64>()
        .expect("Channel id is not a valid id");

    let mut client = Client::builder(token)
        .event_handler(Handler)
        .application_id(application_id)
        .register_songbird()
        .await
        .expect("Error creating client");

    {
        let mut data = client.data.write().await;
        data.insert::<CurrentlyPlayingTrack>(None);
        data.insert::<CurrentVolume>(1.0);
    }

    if let Err(why) = client.start().await {
        println!("Client error: {:?}", why);
    }
}
